@@ -1,11 +1,11 @@
 # -*- coding: utf-8 -*-
 from redis.exceptions import WatchError
 
-from cacheops.conf import redis_conn
+from cacheops.conf import redis_client
 from cacheops.utils import get_model_name
 
 
-__all__ = ('invalidate_obj', 'invalidate_model')
+__all__ = ('invalidate_obj', 'invalidate_model', 'invalidate_all')
 
 
 def serialize_scheme(scheme):
@@ -43,7 +43,7 @@ class ConjSchemes(object):
     def load_schemes(self, model):
         model_name = get_model_name(model)
 
-        txn = redis_conn.pipeline()
+        txn = redis_client.pipeline()
         txn.get(self.get_version_key(model))
         txn.smembers(self.get_lookup_key(model_name))
         version, members = txn.execute()
@@ -84,7 +84,7 @@ class ConjSchemes(object):
                 schemes = self.load_schemes(model)
             if new_schemes - schemes:
                 # Write new schemes to redis
-                txn = redis_conn.pipeline()
+                txn = redis_client.pipeline()
                 txn.incr(self.get_version_key(model_name)) # Увеличиваем версию схем
 
                 lookup_key = self.get_lookup_key(model_name)
@@ -96,14 +96,20 @@ class ConjSchemes(object):
                 self.local[model_name].update(new_schemes)
                 # We increment here instead of using incr result from redis,
                 # because even our updated collection could be already obsolete
-                self.versions[model_name] += 1 
+                self.versions[model_name] += 1
 
     def clear(self, model):
         """
         Clears schemes for models
         """
-        redis_conn.delete(self.get_lookup_key(model))
-        redis_conn.incr(self.get_version_key(model))
+        redis_client.delete(self.get_lookup_key(model))
+        redis_client.incr(self.get_version_key(model))
+
+    def clear_all(self):
+        self.local = {}
+        for model_name in self.versions:
+            self.versions[model_name] += 1
+
 
 cache_schemes = ConjSchemes()
 
@@ -112,51 +118,37 @@ def invalidate_from_dict(model, values):
     """
     Invalidates caches that can possibly be influenced by object
     """
-    # Create a list of invalidators from list of schemes and values of object fields
-    schemes = cache_schemes.schemes(model)
-    conjs_keys = [conj_cache_key_from_scheme(model, scheme, values) for scheme in schemes]
+    def _invalidate(pipe):
+        # Create a list of invalidators from list of schemes and values of object fields
+        schemes = cache_schemes.schemes(model)
+        conjs_keys = [conj_cache_key_from_scheme(model, scheme, values) for scheme in schemes]
 
-    # Get a union of all cache keys registered in invalidators
-    # Get schemes version at the same time, hoping it's unchanged
-    version_key = cache_schemes.get_version_key(model)
-    pipe = redis_conn.pipeline(transaction=False)
-    # Optimistic locking: we hope schemes and invalidators won't change while we remove them
-    # Ignoring this could lead to cache key hanging with it's invalidator removed
-    # HACK: fix for strange WATCH handling in redis-py 2.4.6+
-    if hasattr(pipe, 'pipeline_execute_command'):
-        pipe.pipeline_execute_command('watch', version_key, *conjs_keys)
-    else:
+        # Optimistic locking: we hope schemes and invalidators won't change while we remove them
+        # Ignoring this could lead to cache key hanging with it's invalidator removed
+        version_key = cache_schemes.get_version_key(model)
         pipe.watch(version_key, *conjs_keys)
-    pipe.get(version_key)
-    pipe.sunion(conjs_keys)
-    _, version, cache_keys = pipe.execute()
 
-    # Check if our version if schemes for model is obsolete, update them and redo if needed
-    # This shouldn't be happen too often once schemes are filled a bit
-    if version is None:
-        version = 0
-    if int(version) != cache_schemes.version(model):
-        redis_conn.unwatch()
-        cache_schemes.load_schemes(model)
-        invalidate_from_dict(model, values)
+        # Check if our version of schemes for model is obsolete, update them and redo if needed
+        # This shouldn't be happen too often once schemes are filled a bit
+        version = pipe.get(version_key)
+        if int(version or 0) != cache_schemes.version(model):
+            cache_schemes.load_schemes(model)
+            pipe.reset()
+            raise WatchError() # redo
 
-    elif cache_keys or conjs_keys:
-        # `conjs_keys` are keys of sets containing `cache_keys` we are going to delete,
-        # so we'll remove them too.
-        # NOTE: There could be some other invalidators not matched with current object,
-        #       which reference cache keys we delete, they will be hanging out for a while.
-        try:
-            txn = redis_conn.pipeline()
-            txn.delete(*(list(cache_keys) + conjs_keys))
-            txn.execute()
-        except WatchError:
-            # Optimistic locking failed: schemes or one of invalidator sets changed.
-            # Just redo everything.
-            # NOTE: Scipting will be a cool alternative to optimistic locking
-            invalidate_from_dict(model, values)
+        # Get a union of all cache keys registered in invalidators
+        cache_keys = pipe.sunion(conjs_keys)
+        if cache_keys or conjs_keys:
+            # `conjs_keys` are keys of sets containing `cache_keys` we are going to delete,
+            # so we'll remove them too.
+            # NOTE: There could be some other invalidators not matched with current object,
+            #       which reference cache keys we delete, they will be hanging out for a while.
+            # A transation will fail if schemes or any of invalidator sets is changed in between,
+            # in that case we redo the whole thing
+            pipe.multi()
+            pipe.delete(*(list(cache_keys) + conjs_keys))
 
-    else:
-        redis_conn.unwatch()
+    redis_client.transaction(_invalidate)
 
 
 def invalidate_obj(obj):
@@ -169,15 +161,20 @@ def invalidate_obj(obj):
 def invalidate_model(model):
     """
     Invalidates all caches for given model.
-    NOTE: This is a heavy artilery which uses redis KEYS request, 
+    NOTE: This is a heavy artilery which uses redis KEYS request,
           which could be relatively slow on large datasets.
     """
-    conjs_keys = redis_conn.keys('conj:%s:*' % get_model_name(model))
+    conjs_keys = redis_client.keys('conj:%s:*' % get_model_name(model))
     if isinstance(conjs_keys, str):
         conjs_keys = conjs_keys.split()
 
     if conjs_keys:
-        cache_keys = redis_conn.sunion(conjs_keys)
-        redis_conn.delete(*(list(cache_keys) + conjs_keys))
+        cache_keys = redis_client.sunion(conjs_keys)
+        redis_client.delete(*(list(cache_keys) + conjs_keys))
 
+    # BUG: a race bug here, ignoring since invalidate_model() is not for hot production use
     cache_schemes.clear(model)
+
+def invalidate_all():
+    redis_client.flushdb()
+    cache_schemes.clear_all()
